/*-
 * Copyright (C) @BABOLO  2002 Dec 13
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 13\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: bio_babolo.c,v 1.1 2002/12/14 01:00:39 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "bio_babolo.h"

struct bio_ctrl *
bio_open(int fd, size_t bl) {
    struct bio_ctrl *r;
    struct stat sb;

    r = malloc(sizeof(struct bio_ctrl));
    if  (r == NULL) return(NULL);
    r->fd = fd;
    r->beg = 0;
    r->ep = r->cp = 0;
    if  (0 != fstat(r->fd, &sb)) return(NULL);
    r->mm = (S_ISREG(sb.st_mode) || S_ISDIR(sb.st_mode)) ? 1 : 0;
    r->bs = (sb.st_blksize <= 0) ? 8192 : sb.st_blksize;
    if  (bl > r->bs) r->bs *= (r->bs / bl) + 1;
    r->win = r->bs * 2; /* XXXX       */
    if  (!r->mm && (r->base = malloc(r->win)) == NULL) return(NULL);
    return(r);
};

int
bio_close(struct bio_ctrl *r) {
    int i = 0;

    if  (r->mm) {
        if  (r->base != NULL) i = munmap(r->base, r->win);
    } else free(r->base);
    r->base = NULL;
    r->ep = r->cp = 0;
    i |= close(r->fd);
    r->fd = -1;
    return(i);
};

void *
bio_get(struct bio_ctrl *r, size_t demand) {
    void *m;

    if  (r->cp + demand > r->ep) {
        if  (r->mm) {
            struct stat sb;
            if  (r->base != NULL && munmap(r->base, r->win) < 0) return(NULL);
            r->beg += r->cp;
            r->cp = 0;
            if  (0 != fstat(r->fd, &sb)) return(NULL);
            r->base = mmap(NULL, r->win, PROT_READ, MAP_NOSYNC | MAP_PRIVATE, r->fd, r->beg);
            if  (r->base == MAP_FAILED) return(NULL);
            r->ep = ((sb.st_size - r->beg) > r->win) ? r->win : sb.st_size - r->beg;
        } else {
            ssize_t l = 1;
            if  (r->cp < r->ep) {
                void *v = (void*)(((u_long)r->base & ~0x8) | (((u_long)r->base + r->cp) & 0x7));
                if  (v < r->base) v += 0x8;
                if  (v < r->base + r->cp) {
                    bcopy(r->base + r->cp, v, r->ep - r->cp);
                    r->ep -= r->base + r->cp - v;
                    r->cp = v - r->base;
            }   }
            while (l > 0 && r->cp + demand > r->ep) {
                l = read(r->fd, r->base + r->ep, r->win - r->ep);
                r->ep += l;
    }   }   }
    m = r->base + r->cp;
    r->cp += demand;
    if  (r->cp > r->ep) return(NULL);
    return(m);
};
