/*-
 * Copyright (C) @BABOLO  2002 Dec 08
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 08\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: informat.c,v 1.2 2002/12/14 01:00:39 babolo Exp $";
#endif /* not lint */

#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <err.h>
#include "binfout.h"

#define C_er 0x80000000 /*               */
#define C_cc 0x40000000 /*           */
#define C_tc 0x20000000 /*             */
#define C_tn 0x10000000 /*     */
#define C_tl  0x8000000 /*       */
#define C_tL  0x4000000 /*      */
#define C_ty  0x2000000 /*            */
#define C_state    0xFF /*          */

enum classes {sc, bs, dg, hx, hX, cm, el, en, er, xx};
/* sc :
   dg 0-9
   hx a-f
   hX A-F
   cm #
   el \n
   en ~
   er
   xx
 */

enum states {c0, c1, c2, c3, co, ci, cz, ex};
/* c0 -    ()
   c1 -  :
   c2 - 
   c3 - 
   co - skip comment in c0
   ci - skip comment in c1
   cz - skip comment in c2
   ex
 */

struct formades *
informat(u_char ** const aform) {
#define format (*aform)
    int c = 0;
    enum states state;
    u_int32_t control = 0;
    struct formades *inf;
    size_t i;

    const enum classes class[256] =
    {er,er,er,er ,er,er,er,er ,er,bs,el,er ,er,el,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,bs,er,er,cm ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,dg,dg,dg,dg ,dg,dg,dg,dg ,dg,dg,sc,er ,er,er,er,er

    ,er,hx,hx,hx ,hx,hx,hx,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,hx,hx,hx ,hx,hx,hx,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,en,er

    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er

    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    ,er,er,er,er ,er,er,er,er ,er,er,er,er ,er,er,er,er
    };

    const u_int32_t automa[xx][ex] =
    /*     c0       c1       c2            c3  co  ci  cz */
    {{C_er|ex, C_tc|c2, C_er|ex,      C_er|ex, co, ci, cz}/*sc*/
    ,{     c0,      c1,      c2,      C_er|ex, co, ci, cz}/*bs*/
    ,{C_cc|c1, C_cc|c1, C_tn|c3, C_tn|C_ty|c0, co, ci, cz}/*dg*/
    ,{C_er|ex, C_er|ex, C_tl|c3, C_tl|C_ty|c0, co, ci, cz}/*hx*/
    ,{C_er|ex, C_er|ex, C_tL|c3, C_tL|C_ty|c0, co, ci, cz}/*hX*/
    ,{     co,      ci,      cz,      C_er|ex, co, ci, cz}/*cm*/
    ,{     c0,      c1,      c2,      C_er|ex, c0, c1, c2}/*el*/
    ,{     ex,      ex,      ex,      C_er|ex, co, ci, cz}/*en*/
    ,{C_er|ex, C_er|ex, C_er|ex,      C_er|ex, co, ci, cz}/*er*/
    };

    inf = malloc(sizeof(struct formades));
    inf->max = FORMSIZE;
    inf->cur = -1;
    for (state = c0; state < ex; format++) {
        control = automa[class[*format]][state];
        if  (v_flag > 1) fprintf(stderr, "%08X %02X c=%d\n", control, *format, c);
        if  (control & C_er) exit(EX_DATAERR);
        if  (control & C_cc) {
            if (++(inf->cur) >= inf->max) {
                if  (v_flag >= 0) fprintf(stderr, "Max format %d overloaded\n", inf->max);
                exit(EX_DATAERR);
            }
            inf->form[inf->cur].ll = (*format - '0') << 3;
        }
        if  (control & C_tc) c = 0;
        if  (control & C_tn) c = c << 4 | (*format - '0');
        if  (control & C_tl) c = c << 4 | (*format - 'a' + 10);
        if  (control & C_tL) c = c << 4 | (*format - 'A' + 10);
        if  (control & C_ty) inf->form[inf->cur].ff = c;
        state = control & C_state;
    }
    inf->cur++;
    inf->tot = 0;
    for (i = 0; i < inf->cur; i++) {
        if  (v_flag > 0) fprintf(stderr, " %2d %2d:%02X\n", i, inf->form[i].ll, inf->form[i].ff);
        if  (inf->form[i].ll > 64) {
            if  (v_flag >= 0) fprintf(stderr, "\nmax arg len 64 bit, not %d\n", inf->form[i].ll);
            exit(EX_DATAERR);
        }
        inf->form[i].ff ^= ARCH_PREMASK;
        inf->tot += inf->form[i].ll;
    }
    if  (v_flag > 0)
        fprintf( stderr
               , "total = %d bits (%d Bytes) in %d variables\n"
               , inf->tot, inf->tot >> 3, inf->cur
               );
    if  (inf->tot & 07) {
        if  (v_flag >= 0) fprintf(stderr, "    \n");
        exit(EX_DATAERR);
    }
    return(inf);
}
